/*
 * Copyright (c) 2017, Yuichiro MORIGUCHI
 * All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 * * Redistributions of source code must retain the above copyright notice, 
 *   this list of conditions and the following disclaimer.
 * * Redistributions in binary form must reproduce the above copyright notice, 
 *   this list of conditions and the following disclaimer in the documentation 
 *   and/or other materials provided with the distribution.
 * * Neither the name of the Yuichiro MORIGUCHI nor the names of its contributors 
 *   may be used to endorse or promote products derived from this software 
 *   without specific prior written permission.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED. IN NO EVENT SHALL Yuichiro MORIGUCHI BE LIABLE FOR ANY
 * DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 * LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 * SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */
using System;
using System.IO;
using System.Text;
using System.Collections;
using System.Collections.Generic;
using System.Text.RegularExpressions;
using System.Runtime.Serialization;

namespace Morilib.M4
{
	public class M4Exception : Exception
	{
		public M4Exception() : base()
		{
		}
		public M4Exception(string message) : base(message)
		{
		}
		public M4Exception(string message, Exception inner) : base(message, inner)
		{
		}

		protected M4Exception(SerializationInfo info, StreamingContext context)
        {
        }
	}

	public class M4Object {
		public const int MaxDivert = 10;
		public const int CommentBeginChar = '\u0003';
		public const int CommentEndChar = '\u0004';

		private Dictionary<string, List<string>> macros;
		private string macroName = null;
		private List<string> args = null;
		public TextReader Reader { get; private set; }
		private TextWriter[] divertBuffers;
		private int divnum;
		private int outdivnum;
		public bool DnlFlg { get; set; }
		private string quoteBegin;
		private string quoteEnd;
		private string commentBegin;
		private string commentEnd;
		public ACNode Special { get; private set; }
		public bool SuppressComment { get; private set; }
		public string QuoteBegin
		{
			get
			{
				return quoteBegin;
			}
			set
			{
				quoteBegin = value;
				RefreshACNode();
			}
		}
		public string QuoteEnd
		{
			get
			{
				return quoteEnd;
			}
			set
			{
				quoteEnd = value;
				RefreshACNode();
			}
		}
		public string CommentBegin
		{
			get
			{
				return commentBegin;
			}
			set
			{
				commentBegin = value;
				RefreshACNode();
			}
		}
		public string CommentEnd
		{
			get
			{
				return commentEnd;
			}
			set
			{
				commentEnd = value;
				RefreshACNode();
			}
		}

		public M4Object(M4Object m4obj) {
			macros = m4obj.macros;
			macroName = m4obj.macroName;
			args = m4obj.args;
			Reader = m4obj.Reader;
			divertBuffers = m4obj.divertBuffers;
			divnum = m4obj.divnum;
			outdivnum = m4obj.outdivnum;
			DnlFlg = m4obj.DnlFlg;
			quoteBegin = m4obj.quoteBegin;
			quoteEnd = m4obj.quoteEnd;
			commentBegin = m4obj.commentBegin;
			commentEnd = m4obj.commentEnd;
			Special = m4obj.Special;
		}

		public M4Object(M4Object m4obj, TextReader reader) {
			macros = m4obj.macros;
			macroName = m4obj.macroName;
			args = m4obj.args;
			Reader = reader;
			divertBuffers = m4obj.divertBuffers;
			divnum = m4obj.divnum;
			outdivnum = m4obj.outdivnum;
			DnlFlg = m4obj.DnlFlg;
			quoteBegin = m4obj.quoteBegin;
			quoteEnd = m4obj.quoteEnd;
			commentBegin = m4obj.commentBegin;
			commentEnd = m4obj.commentEnd;
			Special = m4obj.Special;
		}

		public M4Object(TextReader reader, TextWriter writer) {
			macros = new Dictionary<string, List<string>>();
			Reader = reader;
			divertBuffers = new TextWriter[MaxDivert + 1];
			outdivnum = divnum = 0;
			divertBuffers[divnum] = writer;
			DnlFlg = false;
			quoteBegin = "`";
			quoteEnd = "'";
			commentBegin = "#";
			CommentEnd = "\n";
		}

		private void RefreshACNode() {
			Special = new ACNode();
			Special.Add(quoteBegin, '\u0001');
			Special.Add(quoteEnd, '\u0002');
			Special.Add(commentBegin, CommentBeginChar);
			Special.Add(commentEnd, CommentEndChar);
			Special.ComputeFail();
		}

		public void Write(char ch) {
			if(DnlFlg) {
				DnlFlg = ch != '\n';
			} else if(divertBuffers[divnum] != null) {
				if(ch == '\u0001')
				{
					divertBuffers[divnum].Write(QuoteBegin);
				}
				else if(ch == '\u0002')
				{
					divertBuffers[divnum].Write(QuoteEnd);
				}
				else
				{
					divertBuffers[divnum].Write(ch);
				}
			}
		}

		public void Write(string s) {
			string r = s;

			if(!DnlFlg && divertBuffers[divnum] != null) {
				r = r.Replace("\u0001", QuoteBegin);
				r = r.Replace("\u0002", QuoteEnd);
				divertBuffers[divnum].Write(r);
			}
		}

		public void Write(int x) {
			if(!DnlFlg && divertBuffers[divnum] != null) {
				divertBuffers[divnum].Write(x);
			}
		}

		private bool BuiltInMacro(string mname, List<string> lst) {
			List<string> mstk;

			if(mname.Equals("define"))
			{
				if(lst.Count < 1)
				{
					WriteUnevalMacro(mname, lst);
				}
				else
				{
					DefineMacro(lst[0], lst.Count > 1 ? lst[1] : "");
				}
				return true;
			}
			else if(mname.Equals("undefine"))
			{
				if(lst.Count < 1)
				{
					WriteUnevalMacro(mname, lst);
				}
				else if(macros.ContainsKey(lst[0]))
				{
					macros.Remove(lst[0]);
				}
				return true;
			}
			else if(mname.Equals("pushdef"))
			{
				if(lst.Count < 1)
				{
					WriteUnevalMacro(mname, lst);
					return true;
				}
				else if(macros.ContainsKey(lst[0]))
				{
					mstk = macros[lst[0]];
				}
				else
				{
					mstk = new List<string>();
					macros.Add(lst[0], mstk);
				}
				mstk.Add(lst.Count > 1 ? lst[1] : "");
				return true;
			}
			else if(mname.Equals("popdef"))
			{
				if(lst.Count < 1)
				{
					WriteUnevalMacro(mname, lst);
				}
				else if(macros.ContainsKey(lst[0]))
				{
					mstk = macros[lst[0]];
					mstk.RemoveAt(mstk.Count - 1);
					if(mstk.Count == 0)
					{
						macros.Remove(lst[0]);
					}
				}
				return true;
			}
			else if(mname.Equals("defn"))
			{
				if(lst.Count < 1)
				{
					WriteUnevalMacro(mname, lst);
					return true;
				}
				else if(macros.ContainsKey(lst[0]))
				{
					mstk = macros[lst[0]];
					Write(mstk[mstk.Count - 1]);
				}
				return true;
			}
			else if(mname.Equals("ifdef"))
			{
				if(lst.Count < 2)
				{
					WriteUnevalMacro(mname, lst);
				}
				else if(macros.ContainsKey(lst[0]))
				{
					Eval(lst[1]);
				}
				else if(lst.Count > 2)
				{
					Eval(lst[2]);
				}
				return true;
			}
			else if(mname.Equals("ifelse"))
			{
				int cnt = 0;

				for(; lst.Count - cnt >= 3; cnt += 3)
				{
					if(lst[cnt].Equals(lst[cnt + 1]))
					{
						Eval(lst[cnt + 2]);
						return true;
					}
				}
				if(lst.Count - cnt > 0)
				{
					Eval(lst[cnt]);
				}
				return true;
			}
			else if(mname.Equals("shift"))
			{
				for(int i = 1; i < lst.Count; i++)
				{
					Write(lst[i]);
					if(i < lst.Count - 1)
					{
						Write(',');
					}
				}
				return true;
			}
			else if(mname.Equals("dnl"))
			{
				DnlFlg = true;
				return true;
			}
			else if(mname.Equals("divert"))
			{
				if(lst.Count < 1)
				{
					divnum = outdivnum = 0;
				}
				else if(!int.TryParse(lst[0], out outdivnum))
				{
					divnum = outdivnum = MaxDivert;
				}
				else if(outdivnum < 0 || outdivnum > 9)
				{
					divnum = MaxDivert;
				}
				else
				{
					divnum = outdivnum;
				}
				if(divnum < MaxDivert && divertBuffers[divnum] == null)
				{
					divertBuffers[divnum] = new StringWriter();
				}
				return true;
			}
			else if(mname.Equals("undivert"))
			{
				if(lst.Count > 0)
				{
					int div;

					foreach(string arg in lst)
					{
						if(!int.TryParse(arg, out div))
						{
							div = MaxDivert;
						}
						else if(div < 0 || div > 9)
						{
							div = MaxDivert;
						}
						if(div != divnum)
						{
							Write(((StringWriter)divertBuffers[div]).ToString());
							divertBuffers[div] = null;
						}
					}
				}
				else
				{
					for(int i = 1; i < MaxDivert; i++)
					{
						if(i != divnum && divertBuffers[i] != null)
						{
							Write(((StringWriter)divertBuffers[i]).ToString());
							divertBuffers[i] = null;
						}
					}
				}
				return true;
			}
			else if(mname.Equals("changequote"))
			{
				QuoteBegin = lst.Count > 0 ? lst[0] : "`";
				QuoteEnd = lst.Count > 1 ? lst[1] : "'";
				return true;
			}
			else if(mname.Equals("changecom"))
			{
				CommentBegin = lst.Count > 0 ? lst[0] : "#";
				CommentEnd = lst.Count > 1 ? lst[1] : "\n";
				return true;
			}
			else if(mname.Equals("incr") || mname.Equals("decr"))
			{
				if(lst.Count < 1)
				{
					WriteUnevalMacro(mname, lst);
				}
				else
				{
					string sw;
					long bi;

					if(!Int64.TryParse(lst[0], out bi))
					{
						sw = lst[0];
					}
					else if(mname.Equals("incr"))
					{
						sw = (bi + 1).ToString();
					}
					else
					{
						sw = (bi - 1).ToString();
					}
					Write(sw);
				}
				return true;
			}
			else if(mname.Equals("eval"))
			{
				if(lst.Count < 1)
				{
					WriteUnevalMacro(mname, lst);
				}
				else
				{
					Write(M4e.Eval(lst[0]).ToString());
				}
				return true;
			}
			else if(mname.Equals("len"))
			{
				if(lst.Count < 1)
				{
					WriteUnevalMacro(mname, lst);
				}
				else
				{
					Write(lst[0].Length);
				}
				return true;
			}
			else if(mname.Equals("substr"))
			{
				int pos, len;

				if(lst.Count < 3)
				{
					WriteUnevalMacro(mname, lst);
				}
				else
				{
					if(!int.TryParse(lst[1], out pos) || !int.TryParse(lst[2], out len))
					{
						throw new M4Exception();
					}
					else if(pos < 0 && pos >= lst[0].Length)
					{
						throw new M4Exception();
					}
					else if(len < 0 && len + pos >= lst[0].Length)
					{
						throw new M4Exception();
					}
					else
					{
						Write(lst[0].Substring(pos, len));
					}
				}
				return true;
			}
			else if(mname.Equals("translit"))
			{
				if(lst.Count < 3)
				{
					WriteUnevalMacro(mname, lst);
				}
				else
				{
					Translit tbef = new Translit(lst[1]);
					Translit taft = new Translit(lst[2]);
					int index1, ch1;

					for(int i = 0; i < lst[0].Length; i++)
					{
						index1 = tbef.IndexOf(lst[0][i]);
						if(index1 < 0)
						{
							Write(lst[0][i]);
						}
						else if((ch1 = taft.CharAt(index1)) >= 0)
						{
							Write((char)ch1);
						}
					}
				}
				return true;
			}
			else if(mname.Equals("index"))
			{
				if(lst.Count < 2)
				{
					WriteUnevalMacro(mname, lst);
				}
				else
				{
					Write(lst[0].IndexOf(lst[1]).ToString());
				}
				return true;
			}
			else if(mname.Equals("errprint"))
			{
				for(int i = 0; i < lst.Count; i++)
				{
					if(i > 0)
					{
						Console.Error.Write(' ');
					}
					Console.Error.Write(lst[i]);
				}
				Console.Error.WriteLine();
				return true;
			}
			return false;
		}

		public void WriteUnevalMacro(string mname, List<string> lst)
		{
			char cd = '(';

			Write(mname);
			if(lst.Count > 0)
			{
				for(int i = 0; i < lst.Count; i++)
				{
					Write(cd);
					Write(lst[i]);
					cd = ',';
				}
				Write(')');
			}
		}

		private void EvalMacroEvaled(string mname, List<string> lst)
		{
			if(BuiltInMacro(mname, lst))
			{
				return;
			}
			else if(!macros.ContainsKey(mname))
			{
				WriteUnevalMacro(mname, lst);
			}
			else
			{
				M4Object m4obj = new M4Object(this);
				List<string> mstk = macros[mname];
				string mbody;

				m4obj.SuppressComment = true;
				mbody = ScanArgs(mstk[mstk.Count - 1], mname, lst);
				m4obj.Reader = new StringReader(mbody);
				m4obj.macroName = mname;
				m4obj.args = lst;
				M4p.Eval(m4obj);
			}
		}

		private string ScanArgs(string body, string macroName, List<string> args)
		{
			int arg;
			char c;
			StringBuilder b = new StringBuilder(), n;

			for(int i = 0; i < body.Length; i++)
			{
				c = body[i];
				if(c == '$')
				{
					if(++i >= body.Length)
					{
						b.Append(c);
					}
					else if((c = body[i]) >= '0' && c <= '9')
					{
						n = new StringBuilder();
						for(; i < body.Length && (c = body[i]) >= '0' && c <= '9'; i++)
						{
							n.Append(c);
						}
						i--;
						arg = int.Parse(n.ToString());
						if(arg == 0 && macroName != null)
						{
							b.Append(macroName);
						}
						else if(arg > 0 && args != null)
						{
							if(arg - 1 < args.Count)
							{
								b.Append(args[arg - 1]);
							}
						}
					}
					else if("#@*".IndexOf(c) >= 0)
					{
						switch(c)
						{
						case '#':
							b.Append(args.Count);
							break;
						case '*':
							for(int j = 0; j < args.Count; j++)
							{
								b.Append(args[j]);
								if(j < args.Count - 1)
								{
									b.Append(',');
								}
							}
							break;
						case '@':
							for(int j = 0; j < args.Count; j++)
							{
								b.Append(QuoteBegin + args[j] + QuoteEnd);
								if(j < args.Count - 1)
								{
									b.Append(',');
								}
							}
							break;
						}
					}
					else
					{
						b.Append('$');
						b.Append(c);
					}
				}
				else if(c == '\u0001')
				{
					b.Append(QuoteBegin);
				}
				else if(c == '\u0002')
				{
					b.Append(QuoteEnd);
				}
				else
				{
					b.Append(c);
				}
			}
			return b.ToString();
		}

		private void AddAllSplited(List<string> nlst, string text)
		{
			int cnt = 0, par = 0;
			char c;
			StringBuilder b = new StringBuilder();

			for(int i = 0; i < text.Length; i++)
			{
				c = text[i];
				if(cnt > 0)
				{
					if(text.IndexOf(QuoteBegin) == i)
					{
						b.Append(QuoteBegin);
						i += QuoteBegin.Length - 1;
						cnt++;
					}
					else if(text.IndexOf(QuoteEnd) == i)
					{
						b.Append(QuoteEnd);
						i += QuoteEnd.Length - 1;
						cnt--;
					}
					else
					{
						b.Append(c);
					}
				}
				else if(par > 0)
				{
					b.Append(c);
					if(c == '(')
					{
						par++;
					}
					else if(c == ')')
					{
						par--;
					}
				}
				else
				{
					if(text.IndexOf(QuoteBegin) == i)
					{
						b.Append(QuoteBegin);
						i += QuoteBegin.Length - 1;
						cnt++;
					}
					else if(c == '(')
					{
						b.Append(c);
						par++;
					}
					else if(c == ',')
					{
						nlst.Add(b.ToString());
						b = new StringBuilder();
					}
					else
					{
						b.Append(c);
					}
				}
			}
			nlst.Add(b.ToString());
		}

		public void EvalMacro(string mname, List<string> lst)
		{
			List<string> nlst = new List<string>();
			string evd;

			foreach(string str in lst)
			{
				evd = EvalToString(Regex.Replace(str, "^[ \n\t]+", ""));
				AddAllSplited(nlst, evd);
			}
			EvalMacroEvaled(mname, nlst);
		}

		public void Eval(string text)
		{
			M4Object m4obj = new M4Object(this);

			m4obj.SuppressComment = true;
			m4obj.Reader = new StringReader(text);
			M4p.Eval(m4obj);
		}

		public string EvalToString(string text)
		{
			M4Object m4obj = new M4Object(this);
			TextWriter tmp;
			StringWriter sw;

			m4obj.SuppressComment = true;
			m4obj.Reader = new StringReader(text);
			tmp = divertBuffers[divnum];
			m4obj.divertBuffers[divnum] = sw = new StringWriter();
			M4p.Eval(m4obj);
			m4obj.divertBuffers[divnum] = tmp;
			return sw.ToString();
		}

		public void Flush() {
			for(int i = 1; i < MaxDivert; i++)
			{
				if(divertBuffers[i] != null)
				{
					divertBuffers[0].Write(((StringWriter)divertBuffers[i]).ToString());
				}
			}
		}

		public void DefineMacro(string name, string body)
		{
			List<string> mstk;

			if(macros.ContainsKey(name))
			{
				mstk = macros[name];
				mstk[mstk.Count - 1] = body;
			}
			else
			{
				mstk = new List<string>();
				mstk.Add(body);
				macros.Add(name, mstk);
			}
		}
	}

	public class ACNode
	{
		public ACNode Root { get; private set; }
		private Dictionary<int, ACNode> nextNodes = new Dictionary<int, ACNode>();
		public ACNode Fail { get; private set; }
		public int Depth { get; private set; }
		public int QuoteChar { get; private set; }
		public bool IsRoot
		{
			get
			{
				return Root == this;
			}
		}

		public ACNode()
		{
			Root = this;
			Fail = null;
		}

		public ACNode(ACNode root)
		{
			this.Root = root;
			Fail = this;
			QuoteChar = -1;
		}

		public ACNode Transit(int ch)
		{
			return nextNodes.ContainsKey(ch) ? nextNodes[ch] : null;
		}

		public void Add(string str, int quotechar)
		{
			ACNode nd = Root, tn;
			for(int i = 0; i < str.Length; i++)
			{
				if(nd.nextNodes.ContainsKey(str[i]))
				{
					nd = nextNodes[str[i]];
				}
				else
				{
					tn = nd;
					nd = new ACNode(Root);
					nd.Depth = i + 1;
					nd.QuoteChar = -1;
					tn.nextNodes.Add(str[i], nd);
				}
				if(i + 1 == str.Length)
				{
					nd.QuoteChar = quotechar;
				}
			}
		}

		private static bool ComputeFail(ACNode nd, int depth, int execdepth)
		{
			bool ok = false, ta;

			if(depth < execdepth)
			{
				foreach(KeyValuePair<int, ACNode> pair in nd.nextNodes)
				{
					ta = ComputeFail(pair.Value, depth + 1, execdepth);
					ok = ok || ta;
				}
			}
			else if(depth == execdepth)
			{
				foreach(KeyValuePair<int, ACNode> pair in nd.nextNodes)
				{
					ACNode td = nd.Fail;

					while(!td.IsRoot && !td.nextNodes.ContainsKey(pair.Key))
					{
						td = td.Fail;
					}
					pair.Value.Fail = td.IsRoot ? td.Root : td.nextNodes[pair.Key];
					ok = true;
				}
			}
			return ok;
		}

		public void ComputeFail()
		{
			Root.Fail = Root;
			foreach(KeyValuePair<int, ACNode> pair in Root.nextNodes)
			{
				pair.Value.Fail = Root;
			}
			for(int i = 1; ComputeFail(Root, 0, i); i++)
			{
			}
		}
	}

	public class TranslitInterval
	{
		public int BeginChar { get; private set; }
		public int EndChar { get; private set; }
		public int Length
		{
			get
			{
				return EndChar - BeginChar + 1;
			}
		}
		public TranslitInterval(int bg, int en)
		{
			BeginChar = bg;
			EndChar = en;
		}
		public int IndexOf(int ch)
		{
			if(ch >= BeginChar && ch <= EndChar)
			{
				return ch - BeginChar;
			}
			else
			{
				return -1;
			}
		}
		public int CharAt(int index)
		{
			if(index < 0)
			{
				throw new Exception();
			}
			else if(index < Length)
			{
				return BeginChar + index;
			}
			else
			{
				return -1;
			}
		}
	}

	public class Translit
	{
		private List<TranslitInterval> intervals = new List<TranslitInterval>();
		public Translit(string str)
		{
			for(int i = 0; i < str.Length; i++)
			{
				if(i + 2 < str.Length && str[i + 1] == '-')
				{
					if(str[i] <= str[i + 2])
					{
						intervals.Add(new TranslitInterval(str[i], str[i + 2]));
					}
					i += 2;
				}
				else
				{
					intervals.Add(new TranslitInterval(str[i], str[i]));
				}
			}
		}
		public int IndexOf(int ch)
		{
			int ptr = 0, tmp;
			foreach(TranslitInterval inv in intervals)
			{
				tmp = inv.IndexOf(ch);
				if(tmp >= 0)
				{
					return ptr + tmp;
				}
				else
				{
					ptr += inv.Length;
				}
			}
			return -1;
		}
		public int CharAt(int index)
		{
			int ptr = 0, tmp;
			foreach(TranslitInterval inv in intervals)
			{
				tmp = inv.CharAt(index - ptr);
				if(tmp >= 0)
				{
					return tmp;
				}
				else
				{
					ptr += inv.Length;
				}
			}
			return -1;
		}
	}

	public class M4Main
	{
		static void Main(string[] args)
		{
			string all;
			M4Object m4obj = new M4Object(null, Console.Out);
			int cnt;

			for(cnt = 0; cnt < args.Length; cnt++)
			{
				if(args[cnt].StartsWith("-D"))
				{
					string mc = args[cnt].Substring(2, args[cnt].Length - 2);
					int idx = mc.IndexOf('=');
					if(idx < 0)
					{
						m4obj.DefineMacro(mc, "");
					}
					else
					{
						m4obj.DefineMacro(mc.Substring(0, idx),
								mc.Substring(idx + 1, mc.Length - idx - 1));
					}
				}
				else
				{
					break;
				}
			}
			if(cnt < args.Length)
			{
				for(; cnt < args.Length; cnt++)
				{
					using(StreamReader sr = new StreamReader(args[cnt]))
					{
						m4obj = new M4Object(m4obj, sr);
						m4obj = M4p.Eval(m4obj);
						m4obj.Flush();
					}
				}
			}
			else
			{
				while((m4obj = M4p.EvalFromConsole(m4obj)) != null)
				{
					Console.WriteLine();
					m4obj.DnlFlg = false;
					m4obj.Flush();
				}
			}
		    Environment.Exit(0);
		}
	}

}